# -*- coding: utf-8 -*-
"""
Created on Thu Mar  3 10:22:25 2022

@author: bhook

class implemtation of a wordle word.
each word consists of letters that have states from the game

unknown: untested letter
correct: (right letter, right position)
wrong: letter does not exist in the word
misplaced: (right letter, wrong position)

this module will allow the user to load ~all posible wordle solutions, pick a
random word if desired, then iterate through each word eliminating words each
step until a solution is found OR they run out of guesses.

"""

import os
import sys
# set the path so we can find our modules
sys.path.append(os.path.join(os.path.dirname(__file__)))

import wordle_utils
import pandas as pd



class Word:
    """
    Wordle word class that contains stateful letters
    and functions / status variables for signaling game state.
    """

    def __init__(self, word):
        """
        load in the guess word as a string
        init the members.

        Parameters
        ----------
        word : TYPE
            DESCRIPTION.

        Returns
        -------
        None.

        """
        self.word = word
        self.letters = list()
        self.solved = False
        # set all letters to an unknown status initially
        for letter in word:
            self.letters.append(Letter(letter, "unknown"))

    def check_if_solved(self):
        """
        convienence function to check if solved
        (all letters have 'right' status)

        Returns
        -------
        None.

        """
        count = 0
        for letter in self.letters:
            if letter.state == "right":
                count += 1

        if count == 5:
            self.solved = True
            return True
        else:
            self.solved = False
            return False

    def upper(self):
        """
        convenience function to return uppercase
        version of word.word

        Returns
        -------
        TYPE
            DESCRIPTION.

        """
        return self.word.upper()

    def lower(self):
        """
        convenience function to return lowercase
        version of word.word

        Returns
        -------
        TYPE
            DESCRIPTION.

        """
        return self.word.lower()

    def update(self):
        """
        instruct the user for setting each letter state to match
        the latest wordle result on a given attempt.

        Returns
        -------
        None.

        """
        print("For each letter enter: [r/w/m] for right, wrong, or misplaced")
        retry = False
        correct_count = 0
        for letter in self.letters:

            choice = input(f"{letter.letter}:")

            if choice == "r":
                letter.state = "right"
                correct_count += 1

            elif choice == "w":
                # check for duplicate letters that are 'right'
                # if they exist AND they have been marked right
                # change the current letter to 'misplaced' from 'wrong'
                dup_count = 0
                for sub_letter in self.letters:
                    if sub_letter.letter == letter.letter and (
                        sub_letter.state == "right" or sub_letter.state == "misplaced" or sub_letter.state == "unknown"
                    ):
                        dup_count += 1

                if dup_count > 1:
                    print(
                        "duplicate exists elsewhere in the word. 'w' is invalid. I changed your choice to 'misplaced' instead."
                    )
                    letter.state = "misplaced"
                else:
                    letter.state = "wrong"

            elif choice == "m":
                letter.state = "misplaced"

            else:
                print("Invalid choice. The only options are [r/w/m] try again.")
                retry = True
                self.update()
                # break out if the user has retried the input.
                if retry:
                    break

        if correct_count == 5:
            print("SOLVED!")
            self.solved = True


class Letter:
    """
    wordle letter with states
    """

    def __init__(self, letter, state):
        """
        letter is the alphabet characeter
        state refers to its "wordle state"

        Parameters
        ----------
        letter : TYPE
            DESCRIPTION.
        state : TYPE
            DESCRIPTION.

        Returns
        -------
        None.

        """
        self.letter = letter
        self.state = state


def get_remaining_words(word_list, word=Word):
    """
    filter the remaining word list by current letter(s) state
    of the passed Word object. return the new list.

    Parameters
    ----------
    word_list : TYPE
        DESCRIPTION.
    word : TYPE, optional
        DESCRIPTION. The default is Word.

    Returns
    -------
    word_list : TYPE
        DESCRIPTION.

    """

    for index, letter in enumerate(word.letters):

        if letter.state == "right":
            word_list = wordle_utils.letter_at_pos(word_list, letter.letter, index)
        elif letter.state == "wrong":
            word_list = wordle_utils.subwords_not_containing(word_list, [letter.letter])
        elif letter.state == "misplaced":
            word_list = wordle_utils.subwords_containing(word_list, [letter.letter])
            word_list = wordle_utils.letter_not_at_pos(word_list, letter.letter, index)

    return word_list


def main():
    """
    main entry point to the progam

    Returns
    -------
    None.

    """
    print("Loading full wordlist(s)...")

    # read in our full list of english words
    with open(os.path.dirname(__file__) + "/words_alpha.txt", "r") as words_file:
        words = words_file.readlines()

    # keep only the five letter words
    five_letter_words = list()

    for word in words:
        if len(word.strip()) == 5:
            five_letter_words.append(word.strip().lower())

    # Load list of common words to show a relative frequency
    try:
        df = pd.read_excel(
            os.path.dirname(__file__) + "/frequency_list.xlsx",
            sheet_name=2,
            skiprows=1,
            index_col=1,
        )
    except OSError:
        print("failed to find or load the the word frequency list.")
        print("setting the frequency list to an empty dataframe.")
        df = pd.DataFrame()

    df.drop_duplicates(inplace=False)

    # initialize the word list
    word_list = five_letter_words

    print(f"{len(word_list)} five letter words have been loaded. Done.")

    print("Removing prior wordle solutions from the list...")

    with open(os.path.dirname(__file__) + "/wordle_solutions.txt") as solutions_file:
        solutions = solutions_file.readlines()
        solutions = [solution.strip() for solution in solutions]

    for solution in solutions:
        try:
            # filter out blanks/whitespaces
            if len(solution) > 1:
                # remove each prior solution from the word list
                if solution in word_list:
                    word_list.remove(solution)
        except ValueError:
            print(
                f"[{solution}] from the solutions file doesnt appear in the word list."
            )
            print(
                "This shouldn't happen unless the solutions file was modified externally."
            )

    print(f"{len(word_list)} five letter words remain. Done.")

    # you get 6 tries...
    for i in range(6):

        # ask the user for their initial guess
        guess_word = input('Input your word (or type "random") then hit [ENTER]: ')

        # if the user accidentally selects the word index
        # select the word accompanying that number for them:
        if guess_word.isnumeric():
            guess_word = word_list[int(guess_word)]

        guess_word = guess_word.strip()

        if guess_word.lower() == "random":
            guess_word = wordle_utils.get_starting_word(word_list, automated=False)

        guess_word = guess_word.lower()

        assert len(guess_word) == 5, "Passed word must have 5 letters only."

        if guess_word in five_letter_words:

            # initialize wordle word class with guess word str
            word = Word(guess_word)

            print(f'Check "{word.word}" on Wordle then:\n')

            word.update()

            if word.solved:
                print(f"{word.word} is the answer! Congrats!")
                with open(
                    os.path.dirname(__file__) + "/wordle_solutions.txt", "a"
                ) as outfile:
                    # manually strip any newline chars
                    outword = word.word.strip()
                    # then append a windows line ending
                    outfile.write(outword + "\r\n")
                break
            else:
                # find new words from process of elimination()
                word_list = get_remaining_words(word_list, word)

                if len(word_list) > 0:

                    print("\nLatest words to chose from:")

                    for index, word in enumerate(word_list):
                        try:
                            print(
                                index, ". ", word, "\t", df.loc[word, "FREQUENCY"].max()
                            )
                        except KeyError:
                            print(index, ". ", word)
                else:
                    print("Out of words to guess. Double check your inputs.")

        else:
            if guess_word in solutions:
                print(f"[{guess_word}] is a previous solution. Try again. Exit.")
                break
            else:
                print(f"{guess_word} doesn't appear to be in the wordlist. Exit.")
                break


if __name__ == "__main__":
    main()



import os
import sys
# set the path so we can find our modules
sys.path.append(os.path.join(os.path.dirname(__file__)))

import random
import time
import pandas as pd

from selenium import webdriver
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys

# local imports
import wordle_class


def get_starting_word():
    """
    get random starting word from the frequently spoken word list.

    Returns
    -------
    word : TYPE
        DESCRIPTION.

    """

    # get starting word from common words list
    df = pd.read_excel(
        os.path.dirname(__file__) + "/frequency_list.xlsx", sheet_name=2, skiprows=1
    )

    df.drop_duplicates(inplace=True)

    df = df.astype({"LEMMA": "string"})

    df.dropna(inplace=True)

    word_list = df["LEMMA"].to_list()

    five_letter_words = [x.lower() for x in word_list if len(x.strip()) == 5]

    word = five_letter_words[random.randint(0, len(five_letter_words)) - 1]

    word = wordle_class.Word(word)

    return word


def submit_word(driver, word):
    """
    type word into webbrowser using webdriver

    Parameters
    ----------
    driver : TYPE
        DESCRIPTION.
    word : TYPE
        DESCRIPTION.

    Returns
    -------
    None.

    """

    for letter in word.word:
        # random time interval between letters for fun
        time.sleep(random.randint(0, 25) * 0.01)
        webdriver.ActionChains(driver).send_keys(letter).perform()

    time.sleep(0.5)
    # send an enter key to submit word
    webdriver.ActionChains(driver).send_keys(Keys.ENTER).perform()


def check_results(driver, guess_word, rownum):
    """
    scan through the wordle tiles' css state
    Parameters
    ----------
    driver : TYPE
        DESCRIPTION.
    guess_word : TYPE
        DESCRIPTION.
    rownum : TYPE
        DESCRIPTION.

    Returns
    -------
    True or False is the guessed word is valid

    """

    return_code = True

    # each wordle tile is a span. find all spans
    spans = driver.find_elements(By.XPATH, "//span")

    # target the 5 letters on the correct row per guess:
    row = spans[(rownum - 1) * 5 : rownum * 5]

    for index, span in enumerate(row):
        if span.text.lower() == guess_word.word[index]:

            # use the latest css class state to determine
            # the letter state of each guessed letter, then
            # set the wordle Word object letter states
            if "nm-inset-n-gray" in span.get_attribute("class"):
                guess_word.letters[index].state = "wrong"

            elif "nm-inset-yellow" in span.get_attribute("class"):
                guess_word.letters[index].state = "misplaced"

            elif "nm-inset-n-green" in span.get_attribute("class"):
                guess_word.letters[index].state = "right"

            elif "border-red" in span.get_attribute("class"):
                print("invalid wordle word or character")
                return_code = False
            else:
                print("unknown state? race condition?")

    # next handle duplicates
    for letter in guess_word.letters:
        for other_letter in guess_word.letters:
            if (
                letter.letter == other_letter.letter
                and letter.state != other_letter.state
            ):
                if letter.state == "wrong":
                    letter.state = "misplaced"
                elif other_letter.state == "wrong":
                    other_letter.state = "misplaced"

    return return_code


def main(starting_wordle_num):
    """
    main entry point to program
    starts at "starting_wordle_num"
    then iterates through wordles in reverse until
    all wordles have been processed.

    Parameters
    ----------
    starting_wordle_num : TYPE
        DESCRIPTION.

    Returns
    -------
    None.

    """

    for wordle_num in reversed(range(1, starting_wordle_num+1)):

        driver = webdriver.Chrome()
        # implicit 10 second wait added for full DOM/session
        driver.implicitly_wait(10)

        # browse to the wordle archive for a given wordle number
        driver.get(f"https://www.devangthakkar.com/wordle_archive/?{wordle_num}")

        # send an escape key to clear the how-to (if visible)
        webdriver.ActionChains(driver).send_keys(Keys.ESCAPE).perform()

        # print wordle header
        try:
            header = driver.find_element(By.XPATH, "//h1")
            print(header.text)
        except:
            print("no header to find")

        # get starting word returns a Wordle word class
        # the 5 letter guess word will be randomly chosen from
        # frequently spoken words in the word_frequency.xlsx
        starting_word = get_starting_word()

        print(f"starting with {starting_word.upper()}")

        # type word into wordle page
        submit_word(driver, starting_word)

        # check wordle results for a given guess word
        # the wordle Class object letter states are updated
        # accordingly
        while True:
            rc = check_results(driver, starting_word, rownum=1)
            # a -1 return code signifies an invalid wordle
            # in that case, backspace and retry
            if rc is False:
                print(f"{starting_word.word} is invalid! Clearing it out now")
                time.sleep(1)
                for i in range(6):
                    # clear out tiles with slight delay
                    webdriver.ActionChains(driver).send_keys(Keys.BACKSPACE).perform()
                    time.sleep(0.10)

                # get new word and try again
                starting_word = get_starting_word()
                print(f"Guessing {starting_word} from starting retry loop")
                submit_word(driver, starting_word)
            else:
                break

        for letter in starting_word.letters:
            print(f"{letter.letter} - {letter.state}")

        # read full list of possible words for subsequent guessing
        with open(os.path.dirname(__file__) + "/words_alpha.txt", "r") as words_file:
            all_words = words_file.readlines()

        five_letter_words = [x.lower() for x in all_words if len(x.strip()) == 5]

        # get remaining words list with our results of the starting word:
        remaining_word_list = wordle_class.get_remaining_words(
            five_letter_words, starting_word
        )

        print(f"{len(remaining_word_list)} remain for words")

        # init guess word as starting word
        guess_word = starting_word

        # iterate up to five more times while reducing the list
        for i in range(2, 7):

            # check to see if solved
            guess_word.check_if_solved()

            # yes, this is redundant..
            if guess_word.solved:
                print(f"{guess_word.word} is the answer!")
                break

            time.sleep(1)

            # guess random word from remaining
            # TODO: factor in LEMMA score to make better guesses
            if len(remaining_word_list) > 0:
                guess_word = remaining_word_list[
                    random.randint(0, len(remaining_word_list)) - 1
                ]

                print(f"Guessing {guess_word}")

                # convert to wordle word
                guess_word = wordle_class.Word(guess_word)

                # type in word:
                submit_word(driver, guess_word)

                # then drop into check results loop
                while True:
                    rc = check_results(driver, guess_word, rownum=i)
                    if rc is False:
                        print(f"{guess_word.word} is invalid! Clearing it out now")
                        time.sleep(1)
                        for j in range(6):
                            webdriver.ActionChains(driver).send_keys(
                                Keys.BACKSPACE
                            ).perform()
                            time.sleep(0.10)

                        # remove the incorrect word
                        remaining_word_list.remove(guess_word.word)

                        # guess a new word
                        guess_word = remaining_word_list[
                            random.randint(0, len(remaining_word_list)) - 1
                        ]

                        print(f"guessing {guess_word} from main retry loop")

                        # convert to wordle word
                        guess_word = wordle_class.Word(guess_word)
                        submit_word(driver, guess_word)
                    else:
                        break

                time.sleep(1)

                # check if word solved
                if guess_word.check_if_solved():
                    print(f"{guess_word.word} is the answer!")
                    with open("wordlebot_solutions.txt", "w") as solution_file:
                        solution_file.write(f"{wordle_num},{guess_word.word}\n")

                    # break out of guessing for loop
                    # continue main loop for next wordle
                    # after appending answer to text file
                    break

                # remove this guess from the list
                remaining_word_list.remove(guess_word.word)

                remaining_word_list = wordle_class.get_remaining_words(
                    remaining_word_list, guess_word
                )

                print(f"{len(remaining_word_list)} remain(s)")

            elif len(remaining_word_list) == 1:
                print("one word remains! return.")
                break

            else:
                print("out of letters to chose from... hmmm")
                break


if __name__ == "__main__":
    # pick starting wordle
    starting_wordle_num = 271
    main(starting_wordle_num)



# -*- coding: utf-8 -*-
"""
Created on Thu Mar  3 10:33:39 2022

@author: bhook

helper utilities for wordle

"""

import random


def subwords_containing(wordlist, letters=["e", "t", "a", "s"]):
    """
    return a list of words containing the specified list of letters.

    Parameters
    ----------
    wordlist : TYPE
        DESCRIPTION.
    letters : TYPE, optional
        DESCRIPTION. The default is ["e", "t", "a", "s"].

    Returns
    -------
    return_list : TYPE
        DESCRIPTION.

    """

    return_list = list()
    match_count = len(letters)
    count = 0
    for word in wordlist:
        for letter in letters:
            if letter in word:
                count += 1
            else:
                break

        if count == match_count:
            return_list.append(word.strip())
            count = 0
        else:
            count = 0

    return return_list


def subwords_not_containing(wordlist, letters=[]):
    """
    return a list of words NOT containing the specified list of letters

    Parameters
    ----------
    wordlist : TYPE
        DESCRIPTION.
    letters : TYPE, optional
        DESCRIPTION. The default is [].

    Returns
    -------
    TYPE
        DESCRIPTION.

    """
    match_count = len(letters)
    count = 0
    return_list = list()
    for word in wordlist:
        for letter in letters:
            if letter not in word:
                count += 1
            else:
                break

        if count == match_count:
            return_list.append(word.strip())
            count = 0
        else:
            count = 0

    return sorted(return_list)


def letter_at_pos(wordlist, letter, pos):
    """
    return words containing the specifed letter at the specified position.

    Parameters
    ----------
    wordlist : TYPE
        DESCRIPTION.
    letter : TYPE
        DESCRIPTION.
    pos : TYPE
        DESCRIPTION.

    Returns
    -------
    return_list : TYPE
        DESCRIPTION.

    """
    return_list = list()
    for word in wordlist:
        if word[pos] == letter:
            return_list.append(word)

    return return_list


def letter_not_at_pos(wordlist, letter, pos):
    """
    return words NOT containing the specifed letter at the specified position.

    Parameters
    ----------
    wordlist : TYPE
        DESCRIPTION.
    letter : TYPE
        DESCRIPTION.
    pos : TYPE
        DESCRIPTION.

    Returns
    -------
    return_list : TYPE
        DESCRIPTION.

    """
    return_list = list()
    for word in wordlist:
        if word[pos] == letter:
            pass
        else:
            return_list.append(word)

    return return_list


def get_starting_word(five_letter_words, automated=False):
    """
    present the user with random words to start with.
    return the word

    Parameters
    ----------
    five_letter_words : TYPE
        DESCRIPTION.

    Returns
    -------
    TYPE
        DESCRIPTION.

    """

    if automated:
        return five_letter_words[random.randint(0, len(five_letter_words)) - 1]

    else:
        while True:
            starting_word = five_letter_words[
                random.randint(0, len(five_letter_words)) - 1
            ]
            choice = input(
                f"Press [c] to continue with [{starting_word.upper()}] as your starting word, or enter to generate a new one: "
            ).strip()
            if choice.lower() == "c":
                return starting_word.strip()
